using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.CompilerServices;
using System.Diagnostics;
using System.Security.Principal;
using System.ComponentModel;

namespace InfiniTec.DirectoryServices
{
    /// <summary>
    /// Extends the <see cref="ActiveDirectoryEntry"/> with properties unique to 
    /// ActiveDirectory users, groups and contacts
    /// </summary>
    public class ActiveDirectoryPrincipal : ActiveDirectoryEntry
    {
        private Item _BaseItem;

        /// <summary>
        /// Initializes a new instance of the <see cref="ActiveDirectoryPrincipal"/> class.
        /// </summary>
        /// <param name="distinguishedName">The distinguished Name.</param>
        /// <param name="connection">The connection.</param>
        public ActiveDirectoryPrincipal(string distinguishedName, Connection connection)
            :
            base(distinguishedName, connection)
        {
            AddRequiredFields(SchemaMapping.PrincipalSchema);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ActiveDirectoryPrincipal"/> class.
        /// </summary>
        /// <param name="objectId">The object id.</param>
        /// <param name="connection">The connection.</param>
        public ActiveDirectoryPrincipal(Guid objectId, Connection connection)
            :
            base(objectId, connection)
        {
            AddRequiredFields(SchemaMapping.PrincipalSchema);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ActiveDirectoryPrincipal"/> class.
        /// </summary>
        /// <param name="identifier">The identifier.</param>
        /// <param name="connection">The connection.</param>
        public ActiveDirectoryPrincipal(SecurityIdentifier identifier, Connection connection)
            :
            base(string.Format("<Sid={0}>", identifier.Value), connection)
        {
            AddRequiredFields(SchemaMapping.PrincipalSchema);
        }
        /// <summary>
        /// Initializes a new instance of the <see cref="ActiveDirectoryPrincipal"/> class.
        /// </summary>
        /// <param name="baseItem">The base item.</param>
        public ActiveDirectoryPrincipal(Item baseItem)
            : base(baseItem)
        {
            _BaseItem = baseItem;
            AddRequiredFields(SchemaMapping.PrincipalSchema);
        }
        /// <summary>
        /// Gets the name of the account.
        /// </summary>
        /// <value>The name of the account.</value>
        public string NT4AccountName
        {
            get { return GetPropertyValue<string>(Constants.Common.NT4AccountName); }
        }

        /// <summary>
        /// Gets the object sid.
        /// </summary>
        /// <value>The object sid.</value>
        public SecurityIdentifier ObjectSid
        {
            get
            {
                byte[] result;
                result = GetPropertyValue<byte[]>(Constants.Common.ObjectSid);
                return (result != null) ? new SecurityIdentifier(result, 0) : null;
            }
        }

        /// <summary>
        /// Gets the type of the principal.
        /// </summary>
        /// <value>The type of the principal.</value>
        public PrincipalType? PrincipalType
        {
            get
            {
                return (PrincipalType?)GetNullablePropertyValue<int>(Constants.Common.SAMAccountType);
            }
        }
        /// <summary>
        /// Gets a read-only list containing the distinguished names of all direct groups where
        /// the principal is a direct member of.
        /// </summary>
        /// <value>A list containing the distinguished names of all direct groups where
        /// the principal is a direct member of.</value>
        public IList<string> DirectGroupMemberShip
        {
            get
            {
                return GetMultivaluedPropertyValue<string>(Constants.GroupMembership.MemberOf);
            }
        }

        /// <summary>
        /// Gets a read-only list containing the security identifiers 
        /// of all groups (direct and nested) where the principal is a member of.
        /// </summary>
        /// <value>A read-only list containing the security identifiers 
        /// of all groups (direct and nested) where the principal is a member of.</value>
        public IList<SecurityIdentifier> CompleteGroupMembership
        {
            get
            {
                return GetAsSecurityIdentifierList(Constants.GroupMembership.TokenGroups);
            }
        }

        /// <summary>
        /// Gets a read-only list containing the security identifiers 
        /// of all global and universal groups (direct and nested) where the principal 
        /// is a member of.
        /// </summary>
        /// <value>A read-only list containing the security identifiers 
        /// of all global and universal groups (direct and nested) where the principal 
        /// is a member of.</value>
        public IList<SecurityIdentifier> GlobalAndUniversalGroupMembership
        {
            get
            {
                return GetAsSecurityIdentifierList(Constants.GroupMembership.GlobalAndUniversalTokenGroups);
            }
        }

        /// <summary>
        /// Gets a read-only list of security identifiers due to a transitive group membership 
        /// expansion operation on a given user or computer. Token groups cannot be 
        /// retrieved if a Global Catalog is not present to retrieve the transitive 
        /// reverse memberships.
        /// </summary>
        /// <value>The no GC acceptable group membership.</value>
        public IList<SecurityIdentifier> NoGCAcceptableGroupMembership
        {
            get
            {
                return GetAsSecurityIdentifierList(Constants.GroupMembership.NoGCAcceptableTokenGroups);
            }
        }

        private IList<SecurityIdentifier> GetAsSecurityIdentifierList(string propertyName)
        {
            List<SecurityIdentifier> result;
            IList<byte[]> tokenGroups;

            result = new List<SecurityIdentifier>();
            tokenGroups = GetMultivaluedPropertyValue<byte[]>(propertyName);
            if (tokenGroups == null) return result.AsReadOnly();

            foreach (byte[] group in tokenGroups)
            {
                result.Add(new SecurityIdentifier(group, 0));
            }
            return result.AsReadOnly();
        }

        /// <summary>
        /// Translates the sids (from the CompleteGroupMembership property for example) to a 
        /// more readable format
        /// </summary>
        /// <param name="sidsToTranslate">The sids to translate.</param>
        /// <param name="targetFormat">The target format.</param>
        /// <returns>The tranlated values</returns>
        public TranslationResult[] TranslateSids(IEnumerable<SecurityIdentifier> sidsToTranslate, NameFormat targetFormat)
        {
            Translator translator;
            List<string> sids;

            translator = new Translator(BaseItem.Connection);
            translator.InputFormat = NameFormat.SidOrSidHistory;
            translator.OutputFormat = targetFormat;

            sids = new List<string>();
            foreach (SecurityIdentifier sid in sidsToTranslate)
            {
                sids.Add(sid.Value);
            }
            translator.Translate(sids);

            return translator.Results;
        }

        /// <summary>
        /// Opens the specified directory entry and returns either a <see cref="ActiveDirectoryUser"/> or
        /// <see cref="ActiveDirectoryGroup"/>.
        /// </summary>
        /// <param name="distinguishedName">The distinguished name of the object.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <returns>An <see cref="ActiveDirectoryPrincipal"/>.</returns>
        public static ActiveDirectoryPrincipal Open(string distinguishedName, Connection connection)
        {
            ActiveDirectoryPrincipal item;

            item = new ActiveDirectoryPrincipal(distinguishedName, connection);
            item.Refresh();

            if (item.ObjectClass.Contains(Constants.ObjectClasses.Group)) return new ActiveDirectoryGroup(item.BaseItem);
            else if (item.ObjectClass.Contains(Constants.ObjectClasses.User)) return new ActiveDirectoryUser(item.BaseItem);

            throw new InvalidObjectCategoryException("The object doesn't have the proper object categories");
        }

        /// <summary>
        /// Opens the specified directory entry and returns either a <see cref="ActiveDirectoryUser"/> or
        /// <see cref="ActiveDirectoryGroup"/>.
        /// </summary>
        /// <param name="id">The Active Directory <see cref="System.Guid"/> of the object.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <returns>An <see cref="ActiveDirectoryPrincipal"/>.</returns>
        public static ActiveDirectoryPrincipal Open(Guid id, Connection connection)
        {
            return Open(string.Format("<Guid={0}>", id), connection);
        }
        /// <summary>
        /// Opens the specified directory entry and returns either a <see cref="ActiveDirectoryUser"/> or
        /// <see cref="ActiveDirectoryGroup"/>.
        /// </summary>
        /// <param name="sid">The <see cref="SecurityIdentifier"/> of the object.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <returns>An <see cref="ActiveDirectoryPrincipal"/>.</returns>
        public static ActiveDirectoryPrincipal Open(SecurityIdentifier sid, Connection connection)
        {
            return Open(string.Format("<Sid={0}>", sid), connection);
        }
    }
}
