using System;
using System.Collections.Generic;
using System.Text;
using InfiniTec.Threading;
using System.ComponentModel;
using System.Diagnostics;
using System.DirectoryServices.Protocols;
using System.Collections;

namespace InfiniTec.DirectoryServices
{
    /// <summary>
    /// This class abstracts an LDAP-object to level similar to the System.DirectoryServices.DirectoryEntry
    /// class.
    /// </summary>
    public class Item : AsyncOperationObject
    {
        /// <summary>
        /// This event is called when an async refresh operation completes.
        /// </summary>
        public event AsyncCompletedEventHandler RefreshCompleted;
        /// <summary>
        /// This event is called when an async Save operation completes.
        /// </summary>
        public event AsyncCompletedEventHandler SaveCompleted;


        private Connection _Connection;
        private string _DistinguishedName;
        private List<string> _PropertiesToLoad;
        private PropertyCollection _Properties;

        /// <summary>
        /// Gets a list containing the properties to load during a refresh operation.
        /// If left empty, all properties are loaded.
        /// </summary>
        /// <value>The properties to load.</value>
        public IList<string> PropertiesToLoad
        {
            [DebuggerStepThrough]
            get { return _PropertiesToLoad; }
        }


        /// <summary>
        /// Gets or sets the distinguished name.
        /// </summary>
        /// <value>The distinguished name item.</value>
        public string DistinguishedName
        {
            [DebuggerStepThrough]
            get { return _DistinguishedName; }
            [DebuggerStepThrough]
            set { _DistinguishedName = value; }
        }

        /// <summary>
        /// Gets or sets the connection to use.
        /// </summary>
        /// <value>The connection to use.</value>
        public Connection Connection
        {
            [DebuggerStepThrough]
            get { return _Connection; }
            [DebuggerStepThrough]
            set { _Connection = value; }
        }


        /// <summary>
        /// Provides access to the properties of the item.
        /// </summary>
        /// <value>The properties of the item.</value>
        public PropertyCollection Properties
        {
            [DebuggerStepThrough]
            get { return _Properties; }
        }

        private Item()
        {
            _PropertiesToLoad = new List<string>();
            _Properties = new PropertyCollection();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Item"/> class.
        /// </summary>
        /// <param name="distinguishedName">The distinguished name.</param>
        /// <param name="connection">The connection to use.</param>
        public Item(string distinguishedName, Connection connection)
            : this()
        {
            _DistinguishedName = distinguishedName;
            _Connection = connection;
        }

        internal Item(string distinguishedName, SearchResultAttributeCollection attributes, Connection connection)
            :
            this(distinguishedName, connection)
        {
            foreach (DirectoryAttribute attribute in attributes.Values)
            {
                Properties.Add(new RawProperty(attribute));
            }
        }
      

        /// <summary>
        /// This method is called after an async refresh operation has completed.
        /// </summary>
        /// <param name="e">An <see cref="AsyncCompletedEventArgs"/> instance indicating the success
        /// or failure of the refresh operation.</param>
        protected virtual void OnRefreshCompleted(AsyncCompletedEventArgs e)
        {
            if (RefreshCompleted != null) RefreshCompleted(this, e);
        }

        /// <summary>
        /// Synchronously refreshes the properties of the curren item.
        /// </summary>
        public void Refresh()
        {
            using (AsyncOperation<RequestProgress> operation = GetRefreshOperation())
            {
                operation.Run();
                operation.WaitForCompletion();
            }
        }

        /// <summary>
        /// Asynchronously refreshes the properties of this instance.
        /// </summary>
        public void RefreshAsync()
        {
            AsyncOperation<RequestProgress> operation;

            InitiatePendingAsyncOperation();

            operation = GetRefreshOperation();

            RegisterPendingAsyncOperation(operation);
            operation.Completed += RefreshAsyncCompleted;
            operation.Error += RefreshAsyncError;

            operation.Run();
        }


        void RefreshAsyncError(object sender, ExceptionEventArgs e)
        {
            UnregisterPendingAsyncOperation();

            OnRefreshCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        void RefreshAsyncCompleted(object sender, CompletedEventArgs e)
        {
            UnregisterPendingAsyncOperation();

            OnRefreshCompleted(new AsyncCompletedEventArgs(null, e.Status == OperationStatus.Aborted, null));
        }


        /// <summary>
        /// Gets the async operation operation used to refresh the item.
        /// </summary>
        /// <returns>An async operation which can be used to refresh the item.</returns>
        public AsyncOperation<RequestProgress> GetRefreshOperation()
        {
            return new AsyncOperation<RequestProgress>(RefreshInternal());
        }

        private IEnumerable<OperationAction> RefreshInternal()
        {
            SearchRequest request;

            if (Connection == null) throw new InvalidOperationException("The Connection property has not been initialized");

            request = new SearchRequest(DistinguishedName, (string) null, SearchScope.Base, _PropertiesToLoad.ToArray());

            using (RequestOperation<object, SearchResponse> operation = Connection.GetSendRequestOperation<SearchRequest, SearchResponse>(request))
            {
                yield return new NestedOperation(operation, false);

                ParseResponse(operation.Result);
            }
        }

        private void ParseResponse(SearchResponse response)
        {
            SearchResultEntry entry;

            entry = response.Entries[0];

            foreach (DirectoryAttribute attribute in entry.Attributes.Values)
            {
                _Properties.Add(new RawProperty(attribute));
            }
        }

                /// <summary>
        /// This method is called after an async Save operation has completed.
        /// </summary>
        /// <param name="e">An <see cref="AsyncCompletedEventArgs"/> instance indicating the success
        /// or failure of the Save operation.</param>
        protected virtual void OnSaveCompleted(AsyncCompletedEventArgs e)
        {
            if (SaveCompleted != null) SaveCompleted(this, e);
        }

        /// <summary>
        /// Synchronously Saves the properties of the curren item.
        /// </summary>
        /// <param name="lazyCommit">true, if the save operation should return before the changes have been committed to disk. False, otherwise</param>
        public void Save(bool lazyCommit)
        {
            using (AsyncOperation<RequestProgress> operation = GetSaveOperation(lazyCommit))
            {
                operation.Run();
                operation.WaitForCompletion();
            }
        }

        /// <summary>
        /// Asynchronously Saves the properties of this instance.
        /// </summary>
        /// <param name="lazyCommit">true, if the save operation should return before the changes have been committed to disk. False, otherwise</param>
        public void SaveAsync(bool lazyCommit)
        {
            AsyncOperation<RequestProgress> operation;

            InitiatePendingAsyncOperation();

            operation = GetSaveOperation(lazyCommit);

            RegisterPendingAsyncOperation(operation);
            operation.Completed += SaveAsyncCompleted;
            operation.Error += SaveAsyncError;

            operation.Run();
        }


        void SaveAsyncError(object sender, ExceptionEventArgs e)
        {
            UnregisterPendingAsyncOperation();

            OnSaveCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        void SaveAsyncCompleted(object sender, CompletedEventArgs e)
        {
            UnregisterPendingAsyncOperation();

            OnSaveCompleted(new AsyncCompletedEventArgs(null, e.Status == OperationStatus.Aborted, null));
        }


        /// <summary>
        /// Gets the async operation operation used to Save the item.
        /// </summary>
        /// <param name="lazyCommit">true, if the save operation should return before the changes have been committed to disk. False, otherwise</param>
        /// <returns>An async operation which can be used to Save the item.</returns>
        public AsyncOperation<RequestProgress> GetSaveOperation(bool lazyCommit)
        {
            return new AsyncOperation<RequestProgress>(SaveInternal(lazyCommit));
        }

        private IEnumerable<OperationAction> SaveInternal(bool lazyCommit)
        {
            ModifyRequest request;
            ModifyResponse response;

            if (Connection == null) throw new InvalidOperationException("The Connection property has not been initialized");

            request = new ModifyRequest(DistinguishedName);

            request.Modifications.AddRange(Properties.GetModifications());
            if (lazyCommit) request.Controls.Add(new LazyCommitControl());

            if (request.Modifications.Count == 0) yield break;
            
            using (RequestOperation<object, ModifyResponse> operation = Connection.GetSendRequestOperation<ModifyRequest, ModifyResponse>(request))
            {
                yield return new NestedOperation(operation, false);

                response = operation.Result;
            }
            response.ToString();
        }
    }
}
