using System;
using System.Collections.Generic;
using System.Text;
using System.Xml.XPath;
using System.Xml;
using System.DirectoryServices.Protocols;

namespace InfiniTec.DirectoryServices
{
    /// <summary>
    /// Represents a multi-valued property of an Exchange <see cref="Item"/>
    /// </summary>
    /// <typeparam name="T">The type of the property</typeparam>
    public class MultiValuedProperty<T>: Property
    {
        PropertyValueCollection<T> _Values;
        private bool _IsNew;

        /// <summary>
        /// Returns a list containing the elements of the property.
        /// </summary>
        public IList<T> Values
        {
            get { return _Values; }
        }

        /// <summary>
        /// Creates a new instance of this class using the specified namespace and local
        /// name.
        /// </summary>
        /// <param name="name">The full qualified name of the property</param>
        /// <param name="values">The initial values of the property</param>
        public MultiValuedProperty(string name, params T[] values): this(name, true, values)
        {
           
        }

        /// <summary>
        /// Creates a new instance of this class using the specified namespace and local
        /// name.
        /// </summary>
        /// <param name="name">The full qualified name of the property</param>
        /// <param name="isNew">Specifies whether this property is new or already exists on the object</param>
        /// <param name="values">The initial values of the property</param>
        public MultiValuedProperty(string name, bool isNew, params T[] values)
            : this(name, true, isNew, values)
        {
        }

        /// <summary>
        /// Creates a new instance of this class using the specified namespace and local
        /// name.
        /// </summary>
        /// <param name="name">sThe full qualified name of the property</param>
        /// <param name="values">The initial values of the property</param>
        /// <param name="isNew">True, if the property is new on this object. False, otherwise.</param>
        /// <param name="hasChanged">True, if the value is marked as changed. False, otherwise.</param>
        internal MultiValuedProperty(string name, bool hasChanged, bool isNew, params T[] values)
            : base(name)
        {
            _Values = new PropertyValueCollection<T>(this);
            _IsNew = isNew;

            if (values != null) _Values.AddRange(values);

            if (hasChanged) SetChanged();
        }

        /// <summary>
        /// Returns the modifications made to the property as a <see cref="DirectoryAttributeModification"/>
        /// </summary>
        /// <returns>The modifications made to the property as a <see cref="DirectoryAttributeModification"/></returns>
        protected internal override DirectoryAttributeModification GetPropertyModification()
        {
            DirectoryAttributeModification modification;

            modification = new DirectoryAttributeModification();
            modification.Name = Name;
            modification.Operation = (_IsNew) ? DirectoryAttributeOperation.Add : DirectoryAttributeOperation.Replace;

            if (typeof(T) == typeof(byte[]))
            {
                foreach (T value in Values)
                {
                    modification.Add((byte[])((object)value));
                }
            }
            else if (typeof(T) == typeof(Uri))
            {
                foreach (T value in Values)
                {
                    modification.Add((Uri)((object)value));
                }
            }
            else
            {
                foreach (T value in Values)
                {
                    modification.Add(value.ToString());
                }
            }

            return modification;
        }

        /// <summary>
        /// Resets the changed flag.
        /// </summary>
        protected internal override void ResetChangedFlag()
        {
            base.ResetChangedFlag();
            _IsNew = false;
        }

    }
}
