using System;
using System.Collections.Generic;
using System.Text;
using System.Xml.XPath;
using System.DirectoryServices.Protocols;
using System.Diagnostics;

namespace InfiniTec.DirectoryServices
{
    /// <summary>
    /// Represents a single-valued property of an Exchange <see cref="Item"/>
    /// </summary>
    /// <typeparam name="T">The type of the property</typeparam>
    public class Property<T>: Property
    {
        private T _Value;
        private bool _IsNew;

        /// <summary>
        /// Gets or sets the value of the item
        /// </summary>
        public T Value
        {
            get { return _Value; }
            set { _Value = value; SetChanged();  }
        }

        /// <summary>
        /// Creates a new instance of this class using the specified namespace and local
        /// name.
        /// </summary>
        /// <param name="name">The full qualified name of the property</param>
        /// <param name="value">The value of the property</param>
        public Property(string name, T value): this(name, value, true)
        {
        }

        /// <summary>
        /// Creates a new instance of this class using the specified namespace and local
        /// name.
        /// </summary>
        /// <param name="name">The full qualified name of the property</param>
        /// <param name="value">The value of the property</param>
        /// <param name="isNew">Specifies whether this property is new or already exists on the object</param>
        public Property(string name, T value, bool isNew)
            : this(name)
        {
            Value = value;
            _IsNew = isNew;
            SetChanged();
        }

        /// <summary>
        /// Creates a new instance of this class using the specified namespace and local
        /// name.
        /// </summary>
        /// <param name="name">The full qualified name of the property</param>
        public Property(string name): base(name)
        {

        }

        /// <summary>
        /// Returns the modifications made to the property as a <see cref="DirectoryAttributeModification"/>
        /// </summary>
        /// <returns>The modifications made to the property as a <see cref="DirectoryAttributeModification"/></returns>
        protected internal override DirectoryAttributeModification GetPropertyModification()
        {
            DirectoryAttributeModification modification;

            modification = new DirectoryAttributeModification();
            modification.Name = Name;
            modification.Operation = (_IsNew) ? DirectoryAttributeOperation.Add : DirectoryAttributeOperation.Replace;

            if (typeof(T) == typeof(byte[]))
            {
                modification.Add((byte[])((object)Value));
            }
            else if (typeof(T) == typeof(Uri))
            {
                modification.Add((Uri)((object)Value));
            }
            else if (typeof(T).IsClass)
            {
                if (EqualityComparer<T>.Default.Equals(Value, default(T))) modification.Clear();
                else modification.Add(Value.ToString());
            }
            else
            {
                modification.Add(Value.ToString());
            }

            return modification;
        }

        /// <summary>
        /// Resets the changed flag.
        /// </summary>
        protected internal override void ResetChangedFlag()
        {
            base.ResetChangedFlag();
            _IsNew = false;
        }

    }
}
